from typing import Union

from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.database import Database

from database.TweetResult import TweetResult


class TweetResultDAO(object):
    """
    A DAO that lets you save, modify, and retrieve TweetResults from the database.
    """

    def __init__(self, mongoclient: MongoClient, database_name='tweets', collection_name='all_tweets'):
        self.client: MongoClient = mongoclient

        self.database_name = database_name
        self.db: Database = self.client[self.database_name]

        self.collection_name = collection_name
        self.collection: Collection = self.db[self.collection_name]

    def save_one(self, tweetresult: TweetResult):
        """Save a single TweetResult object."""
        self.collection.insert_one(tweetresult.serialize())

    def get_by_id(self, id) -> Union[TweetResult, None]:
        """Return a tweet by ID.

        Returns None if no tweet is found."""
        cursor = self.collection.find({"_id": id})

        if cursor.count() == 0:
            return None

        result = cursor.next()

        cursor.close()

        return TweetResult.deserialize(result)

    def delete_by_id(self, id):
        self.collection.delete_one({"_id": id})



import json
import os


class Config(object):
    """
    This class stores configuration information like config file paths, default values, etc.
    """

    class Defaults:
        """
        This class stores default values for configuration.
        """

        MONGODB_CONNECTION_STRING = "mongodb://localhost:27017/"
        """The connection string for the MongoDB server."""

        API_PORT = 3620
        """The port the HTTP API exposing the Scraper object's functions operates in."""

    @staticmethod
    def example_secrets_json():
        # type: () -> str
        """
        :return: An example of what the secrets.json configuration file should look like.
        """
        return json.dumps({
            'consumer_key': 'ABCDEFG_I_WOULD_LIKE_YOU_TO_REPLACE_ME',
            'consumer_secret': 'ABCDEFG_I_WOULD_LIKE_YOU_TO_REPLACE_ME',
            'access_token': 'ABCDEFG_I_WOULD_LIKE_YOU_TO_REPLACE_ME',
            'access_token_secret': 'ABCDEFG_I_WOULD_LIKE_YOU_TO_REPLACE_ME',
        }, indent=4, sort_keys=True)

    @staticmethod
    def create_example_secrets():
        """Create an example file that serves as `secrets.json` for the user to populate."""

        if not os.path.exists(Config.CONFIG_FOLDER):
            os.mkdir(Config.CONFIG_FOLDER)

        if not os.path.exists(Config.SECRETS_DATAFILE_EXAMPLE_PATH):
            with open(Config.SECRETS_DATAFILE_EXAMPLE_PATH, 'w') as f:
                f.write(Config.example_secrets_json())

    CONFIG_FOLDER = os.path.abspath(os.path.expanduser("~/.twitterfirescraper"))
    """The folder that configuration files, secrets, etc. get stored in."""

    SECRETS_DATAFILE_PATH = os.path.join(CONFIG_FOLDER, "secrets.json")
    """The file that stores Twitter API keys."""

    SECRETS_DATAFILE_EXAMPLE_PATH = os.path.join(CONFIG_FOLDER, "secrets.example.json")
    """The file that serves as an example of the secrets datafile."""

    CONFIG_FILE = os.path.join(CONFIG_FOLDER, "config.json")
    """The file that stores configuration such as rate limiting, API ports, etc."""


class FlaskConfig:
    DEBUG = True
    """Whether or not the app is in DEBUG mode. Should never be TRUE in prod."""

    WEB_PORT = 443
    """The port the website runs on."""

    SSL_CERTIFICATE_PATH = os.path.join(Config.CONFIG_FOLDER, "cert.pem")
    """The location of the SSL certificate."""

    SSL_KEY_PATH = os.path.join(Config.CONFIG_FOLDER, "key.pem")
    """The location of the SSL key."""

    SECRET_KEY = "you will never guess :)"
    """A secret value Flask uses to encrypt session variables. Essentially an encryption key."""

    # If in prod,
    if DEBUG == False:

        SECRET_KEY = os.environ.get("FLASK_SECRET_KEY")

        # If there is no flask secret key,
        if SECRET_KEY is None:
            raise ValueError(
                "No environment variable called `FLASK_SECRET_KEY`! This is required for secure sessions and forms!")


def try_get(object, key):
    """Tries to get the value of a key in an object, and politely notifies you if it cannot find it."""

    if key not in object:
        raise KeyError("Could not find " + key + " inside of " + object + "! Does it exist?")

    return object[key]


class DataConfig:
    """Holds data and filepath configuration information for data files such as state names, hashtags we wish to track,
    terms that indicate fires or disasters, etc."""

    # Folder name that data is stored in.
    DATA_FOLDER_NAME = "data"

    # Path data is stored in.
    DATA_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), DATA_FOLDER_NAME)

    # Disaster hashtag datafile location.
    DISASTER_HASHTAGS_DATA_PATH = os.path.join(DATA_PATH, "DisasterHashtags.yml")

    # Fire hashtag datafile location.
    FIRE_HASHTAGS_DATA_PATH = os.path.join(DATA_PATH, "FireHashtags.yml")

    # US major cities datafile location.
    MAJOR_CITIES_DATA_PATH = os.path.join(DATA_PATH, "MajorCities.yml")

    # US states datafile location.
    US_STATES_DATA_PATH = os.path.join(DATA_PATH, "USStates.yml")

    TWITTER_ACCOUNTS_DATA_PATH = os.path.join(DATA_PATH, "TwitterAccounts.yml")



import unittest

from scraper import Scraper
from twitter import TwitterAuthentication


class TestSimpleScraping(unittest.TestCase):

    def testCanScrapeTerm(self):
        """Tests that scraper can scrape one term."""

        twauth = TwitterAuthentication.autodetect_twitter_auth()

        scraper = Scraper(twitter_authentication=twauth)

        results = scraper.scrape_terms({"fire"}, count=1)

        self.assertIn('fire', results)

        self.assertIsInstance(results['fire'], list)

        self.assertEqual(len(results.keys()), 1)

        self.assertIsInstance(results['fire'][0].get_text(), str)

    def testCanScrapeAccount(self):
        """Tests that scraper can scrape one account."""
        twauth = TwitterAuthentication.autodetect_twitter_auth()

        scraper = Scraper(twitter_authentication=twauth)

        results = scraper.scrape_accounts({"@RedCross"}, count=1)

        self.assertIn('@RedCross', results)

        self.assertEqual(len(results.keys()), 1)

        self.assertIsInstance(results['@RedCross'], list)

        self.assertIsInstance(results['@RedCross'][0].get_text(), str)

    def testCanScrapeMethod(self):
        """Tests that the Scraper's `scrape` method works."""

        twauth = TwitterAuthentication.autodetect_twitter_auth()

        scraper = Scraper(twitter_authentication=twauth)

        results = scraper.scrape(terms={"fire"}, accounts={"@RedCross"})

        self.assertIn('fire', results.keys())
        self.assertIn('@RedCross', results.keys())

        self.assertEqual(len(results.keys()), 2)

        self.assertIsInstance(results['fire'], list)
        self.assertIsInstance(results['@RedCross'], list)

    def _disabled_testLotsOfTweets(self):
        """Tests if the Scraper can retrieve 1000 tweets for one term"""
        twauth = TwitterAuthentication.autodetect_twitter_auth()

        scraper = Scraper(twitter_authentication=twauth)

        results = scraper.scrape(terms={"fire"}, count=1000)

        self.assertIn('fire', results.keys())

        self.assertEqual(len(results.keys()), 1)

        self.assertIsInstance(results['fire'], list)

        self.assertEqual(len(results['fire']), 1000)



# !/usr/bin/env python3

"""
This file starts the web server.
"""
import os
import sys
sys.path.append("..")

from flask import Flask, render_template

from flask_app.forms import ScrapeSingleTermTestForm
from config import FlaskConfig
from scraper import Scraper
from twitter import TwitterAuthentication

current_folder = os.path.abspath(os.path.dirname(__file__))
static_folder = os.path.join(current_folder, 'static')
template_folder = os.path.join(current_folder, 'templates')

app = Flask(__name__, static_url_path='/static', template_folder=template_folder, static_folder=static_folder)

app.debug = FlaskConfig.DEBUG

scraper = Scraper(twitter_authentication=TwitterAuthentication.autodetect_twitter_auth())


@app.route('/')
def index():
    return render_template('index.html')


@app.route("/success")
def success():
    return "Whatever you just tried worked. Congrats :)"


@app.route('/scrape_term', methods=("GET", "POST"))
def scrape_term():
    form = ScrapeSingleTermTestForm()

    tweets = None

    if form.validate_on_submit():
        tweets = scraper.scrape_terms(terms={form.term.data}, count=form.amount.data)

    return render_template('scrape.html', form=form, tweets=tweets)


if __name__ == "__main__":

    app_kwargs = {
        "host": "127.0.0.1",
        "port": FlaskConfig.WEB_PORT,
    }

    if FlaskConfig.DEBUG:

        # Verbose template loading.
        app.config.update({
            "EXPLAIN_TEMPLATE_LOADING": True,
        })

        # Use ad-hoc SSL.
        # This prevents us from having to create an SSL cert in development but still encrypts the connection.
        app_kwargs.update({
            "ssl_context": "adhoc",
        })

    else:  # Not debug mode, production mode.

        # Use SSL cert + key loaded from a file.
        app_kwargs.update({
            "ssl_context": (FlaskConfig.SSL_CERTIFICATE_PATH, FlaskConfig.SSL_KEY_PATH),
        })

    # Update app's config no matter what.
    app.config.update({
        'SECRET_KEY': FlaskConfig.SECRET_KEY,
    })

    app.run(**app_kwargs)



"""
This is a suite of tests that can incur API rate limiting, segregated due to the potentially unwanted affects that can
have.
"""
import os
import sys

# Append parent directory to enhance portability.
sys.path.append((os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))



# coding=utf-8
import unittest

from hashtags import contains_hashtag, extract_hashtags

class TestHashtags(unittest.TestCase):

    def test_extract(self):
        self.assertEqual(extract_hashtags("I love the #beach and #sand!"), {"#beach", "#sand"})

        self.assertEqual(extract_hashtags(u"Pizza is so good! #~# 🍕"), set())

        self.assertEqual(extract_hashtags("Man, Hashtags (#) are # nice # to # use # as # list # columns!"), set())

    def _disabled_test_unicode(self):  # Expect to fail this.
        # TODO add RTL unicode support to hashtag detection.
        self.assertEqual(extract_hashtags(u"مرحبا# to the #world!"), {u'مرحبا#', '#world'})

    def test_simple_contains(self):
        self.assertTrue(contains_hashtag('I enjoy #swimming!', 'swimming'))
        self.assertTrue(contains_hashtag('I enjoy #swimming so much!', 'swimming'))

        self.assertFalse((contains_hashtag("I dislike #coldswimming.", "#swimming")))
        self.assertFalse((contains_hashtag("I dislike #coldswimming.", "#cold")))
        self.assertTrue(contains_hashtag("I dislike #coldswimming.", "#coldswimming"))



import unittest

from tests.test.incur_api_hits.cached_tweets import CachedTweets


class TestNoRetweets(unittest.TestCase):

    def testNoRetweets(self):
        """Tests that scraper can successfully scrape without retweets."""

        results = CachedTweets.tweets_small_no_retweets()

        for term, statuses in results.items():
            for status in statuses:
                self.assertFalse(status.data.retweeted)

                # Author of original tweet and person posting MUST be the same.
                # If not, it is a retweet.
                self.assertEqual(status.data.author.id, status.data.user.id)



"""
This file tests that the scraper fails correctly with invalid credentials.
"""
import unittest

from tweepy import TweepError

from scraper import Scraper
from twitter import TwitterAuthentication


class TestInvalidCredentials(unittest.TestCase):

    def test_invalid(self):
        # This is one way to ideally initialize a Scraper object.
        scraper = Scraper(twitter_authentication=TwitterAuthentication(
            consumer_key="i4m4c0ns4m3r",
            consumer_secret="itsasecret",
            access_token="chuckecheesetokens?",
            access_token_secret="ticketmuncherdoesn'twork"
        ))

        terms = {"#pizza", "pizza", 'cat', '#cat'}

        failure = False

        try:
            results = scraper.scrape_terms(terms=terms, count=3)
        except TweepError as e:
            print("Failed successfully with bogus credentials.")
            failure = True

        if not failure:
            self.fail("Credentials should fail to validate.")


from flask_wtf import FlaskForm
from wtforms import StringField, IntegerField
from wtforms.validators import DataRequired, NumberRange

class ScrapeSingleTermTestForm(FlaskForm):
    term = StringField('term', validators=[DataRequired()])
    amount = IntegerField('amount', validators=[NumberRange(1, 1000)])


#!/usr/bin/env python3

from flask import Flask, request, jsonify, abort
from flask_pymongo import PyMongo

# from flask.ext.pymongo import PyMongo
from config import Config
from scraper import Scraper
from twitter import TwitterAuthentication
from util import jsonify_status_dict

app = Flask(__name__, static_url_path="/static")

app.config['MONGO_DBNAME'] = 'twitterfirescraperapi'  # name of your cluster
app.config['MONGO_URI'] = 'mongodb+srv://trungpham10:!twitterfire1@twitterfirescraperapi-i6mwc.mongodb.net/test?retryWrites=true'
# mongodb username and password here

mongo = PyMongo(app)

scraper = Scraper(twitter_authentication=TwitterAuthentication.autodetect_twitter_auth())


@app.route('/add_new', methods=['GET'])
def add():
    user = mongo.db.users
    user.insert_one({'city': 'New York'})
    return 'New York is added!'


@app.route('/scrape_terms', methods=['GET'])
def scrape_terms():
    # user1 = mongo.db.users

    print('scrape_terms')
    print(request.args)

    count = request.args.get("count")
    if not count:
        abort(400, "'count' is a required URL parameter!")

    try:
        count = int(count)

        if count < 0:
            raise ValueError
    except ValueError:
        abort(400, "'count' should be a valid number!")

    terms = request.args.get("terms")

    print(terms)

    if not terms:
        abort(400, "'terms' is a required URL parameter!")

    terms = terms.split(",")
    if len(terms) <= 0:
        abort(400, "'terms' cannot be an empty list!")

    geocode = request.args.get("geocode")

    results = (scraper.scrape_terms(terms=terms, count=count, geocode=geocode))  # dict object

    # user1.insert(results)

    results = jsonify_status_dict(results)  # json object

    return jsonify(results)
    # return jsonify("You want {} tweets for {}?".format(count, terms))


@app.route('/scrape_accounts', methods=['GET'])
def scrape_accounts():
    count = request.args.get("count")
    if not count:
        abort(400, "'count' is a required URL parameter!")
    try:
        count = int(count)
        if count < 0:
            raise ValueError
    except ValueError:
        abort(400, "'count' should be a valid number!")

    accounts = request.args.get("accounts")
    if not accounts:
        abort(400, "'accounts' is a required URL parameter!")
    accounts = accounts.split(",")
    if len(accounts) <= 0:
        abort(400, "'accounts' cannot be an empty list!")

    results = (scraper.scrape_accounts(accounts=accounts, count=count))

    results = jsonify_status_dict(results)

    return jsonify(results)
    # return jsonify("You want {} tweets for {}?".format(count, term))


@app.route('/scrape', methods=['GET'])
def scrape():
    count = request.args.get("count")
    if not count:
        abort(400, "'count' is a required URL parameter!")
    try:
        count = int(count)
        if count < 0:
            raise ValueError
    except ValueError:
        abort(400, "'count' should be a valid number!")

    terms = request.args.get("terms")
    if not terms:
        abort(400, "'terms' is a required URL parameter!")
    terms = terms.split(",")
    if len(terms) <= 0:
        abort(400, "'terms' cannot be an empty list!")

    accounts = request.args.get("accounts")
    if not accounts:
        abort(400, "'accounts' is a required URL parameter!")
    accounts = accounts.split(",")
    if len(accounts) <= 0:
        abort(400, "'accounts' cannot be an empty list!")

    geocode = request.args.get("geocode")

    results = (scraper.scrape(terms=terms, accounts=accounts, count=count, geocode=geocode))

    results = jsonify_status_dict(results)

    return jsonify(results)
    # return jsonify("You want {} tweets for {}?".format(count, term))


@app.route('/info', methods=['GET'])
def info():  # function: check webapi is running or not
    return "twitter-fire-scraper-webapi"


@app.route('/')
def index():
    return jsonify({'info': 'Welcome. This is the twitter-fire-scraper-webapi. This is a Web API, and not meant to be viewed in your browser.'})


@app.route('/add/<int:x>/<int:y>', methods=['GET'])
def add_numbers(x, y):
    return str(x + y)


if __name__ == "__main__":

    app.run(host="127.0.0.1", port=Config.Defaults.API_PORT, debug=True)



import unittest

from pymongo import MongoClient

from config import Config
from database.TweetResult import TweetResult, ERelevancy
from database.TweetResultDAO import TweetResultDAO
from tests.test.incur_api_hits.cached_tweets import CachedTweets


class TestDaoSimple(unittest.TestCase):

    def setUp(self) -> None:
        # connect to localhost
        self.client = MongoClient(Config.Defaults.MONGODB_CONNECTION_STRING)

        # make a DAO with our connection
        self.dao = TweetResultDAO(self.client, collection_name=self.__class__.__name__)

    def tearDown(self) -> None:
        self.client.close()

    def testRealData(self):
        """Tests that the DAO can successfully insert and delete a few real records."""
        self.assertTrue(True)  # TODO

        # small amount of tweets
        tweetResultDict = CachedTweets.tweets_small()


        # save all of them
        for term, tweetResults in tweetResultDict.items():

            for tweetResult in tweetResults:

                # Make sure it doesn't already exist if we're running this SUPER fast or twice.
                if self.dao.get_by_id(tweetResult.get_id()) is not None:
                    self.dao.delete_by_id(tweetResult.get_id())

                self.dao.save_one(tweetResult)

        # using our local copy, retrieve them and compare them
        for term, tweetResults in tweetResultDict.items():

            for tweetResult in tweetResults:

                database_tweetResult = self.dao.get_by_id(tweetResult.get_id())

                # for brevity, compare their serialized forms.
                self.assertEqual(tweetResult.serialize(), database_tweetResult.serialize())

    def testMockData(self):
        """Tests that the DAO can successfully insert and delete a single mock record."""

        # delete if it somehow exists
        if self.dao.get_by_id('test_id') is not None:
            self.dao.delete_by_id('test_id')

        # mock tweetresult
        tweetResult = TweetResult(
            data={"id": "test_id",
                  "full_text": "hello!"},
            tags=["test", "deleteme"],
            relevancy=ERelevancy.IRRELEVANT)

        # save a tweetresult
        self.dao.save_one(tweetResult)

        # delete our local copy!
        del tweetResult

        # deserialize it from db
        database_result = self.dao.get_by_id("test_id")

        # It should exist
        self.assertNotEqual(database_result, None)

        # it should be of type TweetResult
        self.assertIsInstance(database_result, TweetResult)

        # it should have its tags,
        self.assertIn("test", database_result.tags)
        self.assertIn("deleteme", database_result.tags)

        # it should have its ERelevancy
        self.assertEqual(ERelevancy.IRRELEVANT, database_result.relevancy)

        # delete it now
        self.dao.delete_by_id(database_result.get_id())

        # it should be None since it's gone
        self.assertEqual(self.dao.get_by_id("test_id"), None)



"""
A set of classes that allows access to a MongoDB database to save, update, and retrieve TweetResult objects.
"""
import os
import sys

# Append parent directory to enhance portability.
sys.path.append((os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))



import sys
from enum import Enum

sys.path.append("..")

from tweepy import Status
from typing import Dict, List, Union

from util import status_from_dict


class ERelevancy(Enum):
    """An enumeration that represents if something is uncategorized, relevant, or irrelevant."""
    UNCATEGORIZED = "UNCATEGORIZED"
    RELEVANT = "RELEVANT"
    IRRELEVANT = "IRRELEVANT"


class TweetResult(object):
    """An object representing a single Tweet along with metadata such as tags and relevancy."""

    def __init__(self,
                 data: Union[Status, dict],
                 tags: List[str],
                 relevancy: ERelevancy):

        # Convert a dict to a Status automatically
        if isinstance(data, dict):
            data = status_from_dict(data)

        if not isinstance(data, Status):
            raise ValueError("You must pass a Status object to the TweetResult constructor!")

        self.data = data
        self.tags = tags
        self.relevancy = relevancy

    def get_json(self) -> dict:
        """The JSON value that the Twitter API returns for this tweet."""
        return self.data._json

    def get_id(self):
        """The ID of a Tweet. This is unique."""
        return self.data.id

    def get_text(self):
        """The content of a tweet.

        Returns either "full_text" or "text", favoring the former as it is longer."""

        if hasattr(self.data, "full_text"):
            return self.data.full_text

        elif hasattr(self.data, 'text'):
            return self.data.text

        else:
            raise KeyError("Tweet has no text!")

    def serialize(self):
        """Serialize this TweetResult object to a JSON-like object that can be saved to a MongoDB database"""

        if type(self.tags) is set:
            self.tags = list(self.tags)

        return {
            # Base attributes, these are needed to fully reconstruct the object.
            'data': self.get_json(),
            'tags': self.tags,
            'relevancy': self.relevancy.value,

            # Derived attributes, these are inferred or created from the base attributes.
            '_id': self.get_id(),
        }

    @classmethod
    def deserialize(cls, dict: Dict):
        """Deserialize a JSON-like object to a TweetResult. Like the opposite of serializing."""

        status = status_from_dict(dict['data'])

        return TweetResult(data=status,
                           tags=dict['tags'],
                           relevancy=ERelevancy(dict['relevancy']))



import csv
import os
from typing import Dict, List, Set

import tweepy
from pymongo import MongoClient
from tweepy import Status

from database.TweetResult import TweetResult, ERelevancy
from twitter import TwitterAuthentication
from util import merge_status_dict, save_statuses_dict_to_mongodb


class Scraper:
    CSV_DELIMITER = '\t'

    def __init__(self, twitter_authentication=None):
        # type: (Scraper, TwitterAuthentication) -> None

        # They did not pass in any authentication. Attempt to auto-detect it.
        if not twitter_authentication:
            self.twitter_authentication = TwitterAuthentication.autodetect_twitter_auth()
        else:  # They passed us a TwitterAuthentication object
            self.twitter_authentication = twitter_authentication

        # Tweepy API object. Can make API calls.
        self.api = tweepy.API(self.twitter_authentication.oauth_handler,
                              wait_on_rate_limit=True, wait_on_rate_limit_notify=True)

        # Default amount of tweets to retrieve.
        self.default_count = 10

        # Default accounts to scrape.
        # self.accounts = accounts in data/TwitterAccounts.yml

    def scrape_terms(self,
                     terms: Set[str],
                     count: int = None,
                     geocode: str = None,
                     include_retweets: bool = True) -> Dict[str, List[TweetResult]]:
        """
        Term-scraping method. Can scrape a set of terms.

        A term is either a hashtag or a piece of text.

        :param geocode: Geographical area to search in. Can be blank.
        :param terms:  List of terms to search for.
        :param count: Maximum tweets to return per search term.
        :param include_retweets: Should retweets be included?
        :return: A dictionary containing {'search-term': List[TweetResult]} pairs.
        """

        if not count:
            count = self.default_count

        results: Dict[str, List[TweetResult]] = {}

        # For each search term,
        for search_term in terms:

            query = search_term

            if not include_retweets:
                query += ' -filter:retweets'
                query += ' -filter:nativeretweets'

            # Make a cursor that can get tweets.
            cursor = tweepy.Cursor(self.api.search, q=query, geocode=geocode, tweet_mode='extended')

            # For each result of a search term,
            for status in cursor.items(count):  # type: Status

                if search_term not in results:
                    results[search_term] = list()

                result = TweetResult(data=status, tags=[search_term], relevancy=ERelevancy.UNCATEGORIZED)

                # Add the status to a particular search term.
                results[search_term].append(result)

        return results

    def scrape_accounts(self,
                        accounts: Set[str],
                        count: int = None) -> Dict[str, List[TweetResult]]:
        """
        Account-scraping method. Can scrape a set of accounts.

        An account is either a screenname or @screenname, i.e. both 'Dude123' and '@Dude123' are valid.

        :param accounts: List of accounts to search in.
        :param count: Maximum tweets to return per account.
        :return: A dictionary containing {'@AccountName': List[TweetResult]} pairs.
        """

        if not count:
            count = self.default_count

        results = dict()

        for account in accounts:
            statuses: List[Status] = self.api.user_timeline(screen_name=account, count=count)

            if account not in results:
                results[account] = list()

            for status in statuses:
                result = TweetResult(data=status, tags=[account], relevancy=ERelevancy.UNCATEGORIZED)

                results[account].append(result)

                return results

    def scrape(self, terms: Set[str] = None,
               accounts: Set[str] = None,
               count: int = None,
               geocode: str = None,
               include_retweets: bool = True) -> Dict[str, List[TweetResult]]:
        """
        General-purpose scraping method. Can scrape search terms, and accounts.

        :param geocode: Geographical area to search in. Can be blank.
        :param terms:  List of terms to search for.
        :param accounts: List of account names to search.
        :param count: Maximum tweets to return per search term.
        :param include_retweets: Should retweets be included?
        :return: A dictionary containing {'search-term': List[Status]} pairs.

        Examples:
            >>> self.scrape(geocode="41.8297855,-87.666775,50mi", terms={"pizza", "waffles"}, count=3)
            {'pizza': {TweetResult, TweetResult, TweetResult},
            'waffles': {TweetResult, TweetResult}}
        """
        if not count:
            count = self.default_count

        if (not terms) and (not accounts):
            raise ValueError("No terms or accounts specified.")

        results = dict()

        if terms:
            terms_results = self.scrape_terms(terms=terms, count=count, geocode=geocode,
                                              include_retweets=include_retweets)

            results = merge_status_dict(results, terms_results)

        if accounts:
            accounts_results = self.scrape_accounts(accounts=accounts, count=count)

            results = merge_status_dict(results, accounts_results)

        return results

    def scrape_and_save(self,
                        terms: Set[str] = None,
                        accounts: Set[str] = None,
                        count: int = None,
                        geocode: str = None,
                        dbname='scraper_tweets'):

        # First, retrieve search results via scrape
        results = self.scrape(terms=terms, accounts=accounts, count=count, geocode=geocode)

        # Establish connection to the host
        client = MongoClient()

        # Access database where the scraped tweets will be saved
        db = client[dbname]

        # Save all tweets to the database
        save_statuses_dict_to_mongodb(results, db)

        return results

    def save_statusdict_to_csv(self,
                               statusdict: Dict[str, List[TweetResult]],
                               filepath: str,
                               overwrite: bool = False,
                               relevant_column: bool = False):
        """Save a status dict to a CSV file.
        :param statusdict A {str: [TweetResult, TweetResult, ...]} dictionary.
        :param filepath A path to the file to output to.
        :param overwrite Should we overwrite `filepath` if it exists?
        :param relevant_column Should we insert an extra column for hand-categorizing in Excel?
        """

        dirname = os.path.dirname(filepath)

        if not os.path.exists(dirname):
            raise NotADirectoryError("Directory {} does not exist!".format(dirname))

        if os.path.isfile(filepath):
            if not overwrite:
                raise FileExistsError("File at '{}' already exists!".format(filepath))

        fieldnames = [
            'category', 'tweet_id', 'text', 'date', 'retweet_count',

            'geo',

            'coordinates',

            'place_id', 'place_centroid', 'place_country', 'place_country_code', 'place_full_name',
            'place_name', 'place_type'
        ]

        if relevant_column:
            fieldnames.append("relevant")

        with open(filepath, 'w', encoding='utf-16', newline='') as file:
            fileWriter = csv.DictWriter(file, delimiter=Scraper.CSV_DELIMITER, quotechar='"', quoting=csv.QUOTE_ALL,
                                        fieldnames=fieldnames)

            fileWriter.writeheader()

            for keyword, statuses in statusdict.items():

                for status in statuses:

                    data = {
                        "category": keyword,
                        "tweet_id": status,
                        "text": status.get_text(),
                        "date": status.data.created_at,
                        'retweet_count': status.data.retweet_count,
                    }

                    if relevant_column:
                        data.update({
                            'relevant': "UNCATEGORIZED"
                        })

                    if status.data.geo:
                        data.update({
                            "geo": ','.join(str(x) for x in status.data.geo['coordinates']),
                        })

                    if status.data.coordinates:
                        data.update({
                            'coordinates': ','.join(str(x) for x in status.data.coordinates['coordinates']),
                        })

                    # If the status has a place, then add its data!
                    if status.data.place:
                        data.update({
                            'place_id': status.data.place.id,
                            # We have to use the API for this one to look it up.
                            'place_centroid': ','.join(str(x) for x in self.api.geo_id(status.data.place.id).centroid),
                            'place_country': status.data.place.country,
                            'place_country_code': status.data.place.country_code,
                            'place_full_name': status.data.place.full_name,
                            'place_name': status.data.place.name,
                            'place_type': status.data.place.place_type,
                        })

                    # Write all the data we've collected so far.
                    fileWriter.writerow(data)

        return filepath
